"""Script to be invoked to run the test suite.

It sets up the Django environment, creates the test database and runs the tests.
"""

import argparse
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


def parse_args(argv=None):
    """Parse command-line arguments for the test runner."""
    parser = argparse.ArgumentParser(description="Run the django-telegram-app test suite.")

    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=1,
        help=(
            "Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output (default: 1)."
        ),
    )
    parser.add_argument(
        "-q",
        "--buffer",
        action="store_true",
        help=("Buffer stdout/stderr during tests. Only show output for failing/erroring tests."),
    )
    parser.add_argument(
        "test_labels",
        nargs="*",
        default=["tests"],
        help="Specific test labels to run (default: 'tests').",
    )

    return parser.parse_args(argv)


def main(argv=None):
    """Run the test suite."""
    args = parse_args(argv)
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.testapps.settings"
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner(verbosity=args.verbosity, buffer=args.buffer)
    failures = test_runner.run_tests(args.test_labels)
    sys.exit(bool(failures))


if __name__ == "__main__":
    main()



"""Telegram app."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from django_telegram_app.models import AbstractTelegramSettings


def get_telegram_settings_model() -> type[AbstractTelegramSettings]:
    """Return the TelegramSettings model that is active in this project.

    If there is no custom model, return the default TelegramSettings model.
    """
    from django_telegram_app.models import AbstractTelegramSettings

    try:
        telegram_settings_model = settings.TELEGRAM_SETTINGS_MODEL
    except AttributeError:
        from django_telegram_app.models import TelegramSettings

        return TelegramSettings

    try:
        return cast(type[AbstractTelegramSettings], django_apps.get_model(telegram_settings_model, require_ready=False))
    except ValueError as exc:
        raise ImproperlyConfigured("TELEGRAM_SETTINGS_MODEL must be of the form 'app_label.model_name'") from exc
    except LookupError as exc:
        raise ImproperlyConfigured(
            f"TELEGRAM_SETTINGS_MODEL refers to model '{telegram_settings_model}' that has not been installed"
        ) from exc



"""Bot module for the Telegram app."""

from __future__ import annotations

import functools
import pkgutil
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING

from django.apps import apps
from django.conf import settings

if TYPE_CHECKING:
    from django_telegram_app.bot.base import BaseBotCommand
    from django_telegram_app.models import AbstractTelegramSettings


def find_commands(telegrambot_dir: Path):
    """Return a list of all the command names that are available for the provided telegrambot path."""
    command_dir = telegrambot_dir / "commands"
    return [name for _, name, is_pkg in pkgutil.iter_modules([command_dir]) if not is_pkg and not name.startswith("_")]


def load_command_class(app_name: str, name: str, settings: AbstractTelegramSettings) -> BaseBotCommand:
    """Return the Command class instance for the given command name and application name.

    Allow all errors raised by the import process (ImportError, AttributeError) to propagate.
    """
    module = import_module(f"{app_name}.telegrambot.commands.{name}")
    return module.Command(settings)


@functools.cache
def get_commands():
    """Return a dictionary mapping command names to their callback applications.

    Look for a telegrambot.commands package in each installed application -- if a commands package exists, register all
    commands in that package.

    All user-defined commands from the specified settings module are included.

    The dictionary is in the format {command_name: app_name}. Key-value
    pairs from this dictionary can then be used in calls to
    load_command_class(app_name, command_name)

    The dictionary is cached on the first call and reused on subsequent
    calls.
    """
    commands: dict[str, str] = {}
    if not settings.configured:
        return commands

    for app_config in apps.get_app_configs():
        path = Path(app_config.path) / "telegrambot"
        commands.update({name: app_config.name for name in find_commands(path)})

    return commands



"""Base command module to start telegram commands.

Note:
    This module's command-class does not create an actual CLI command, but can be used by actual commands.
"""

from django.core.management.base import BaseCommand

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.bot.base import BaseBotCommand
from django_telegram_app.bot.bot import handle_update
from django_telegram_app.models import AbstractTelegramSettings


class BaseManagementCommand(BaseCommand):
    """Base command class to start Telegram bot commands.

    Subclasses can override:
        - the `should_run` method to determine if the command should run.
        - the `get_telegram_settings_filter` method to filter telegram settings.
        - the `handle_command` method to customize the update handling
    """

    command: type[BaseBotCommand] | None = None

    def add_arguments(self, parser):
        """Add command arguments."""
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force the command to run, regardless of the should_run outcome.",
        )

    def handle(self, *_args, **options):
        """Start the configured telegram command.

        This method should not be overridden by subclasses.

        Subclasses can override:
            - the `should_run` method to determine if the command should run.
            - the `get_telegram_settings_filter` method to filter telegram settings.
            - the `handle_command` method to customize the update handling
        """
        if not self.command:
            raise ValueError("The attribute `command` must be set.")
        command_text = self.command.get_command_string()

        if not options["force"] and not self.should_run():
            self.stdout.write(self.style.NOTICE(f"Command '{command_text}' skipped as `should_run` returned False."))
            return

        telegram_settings_filter = self.get_telegram_settings_filter()
        for telegram_settings in get_telegram_settings_model().objects.filter(**telegram_settings_filter):
            self.handle_command(telegram_settings, command_text)
            self.stdout.write(self.style.SUCCESS(f"Started the command for {telegram_settings}."))
        else:
            self.stdout.write(self.style.NOTICE("No Telegram-settings found for the given filter. Nothing to do."))

    def should_run(self) -> bool:
        """Determine if the command should run."""
        return True

    def get_telegram_settings_filter(self):
        """Get the filter to apply when retrieving telegram settings.

        Can be overridden by subclasses to filter telegram settings.
        By default, all telegram settings are returned.
        """
        return {}

    def handle_command(self, telegram_settings: AbstractTelegramSettings, command_text: str):
        """Construct a telegram update and handle it.

        Subclasses are encouraged to override this method to customize the update handling.
        (e.g., to add custom data to the update, to localize, etc.)

        Note:
            A minimal update is created with a message containing the command, this update is not persisted.
        """
        update = {"message": {"chat": {"id": telegram_settings.chat_id}, "text": command_text}}
        handle_update(update=update, telegram_settings=telegram_settings)



"""Reusable testcases for Telegram bot app."""

import warnings
from unittest.mock import MagicMock, patch

from django.test.testcases import TestCase
from django.urls import reverse

from django_telegram_app.conf import settings


class TelegramBotTestCase(TestCase):
    """Base test case for Telegram bot tests."""

    @property
    def webhook_url(self):
        """Return the webhook URL."""
        return reverse("webhook")

    @classmethod
    def tearDownClass(cls):
        """Tear down class-level patches."""
        try:
            super().tearDownClass()
        finally:
            patch.stopall()

    def setUp(self):
        """Set up each test."""
        self.fake_bot_post = patch("django_telegram_app.bot.bot.post", MagicMock()).start()

    def click_on_text(self, text: str, verify: bool = True):
        """Simulate a click on the specified text button."""
        warnings.warn(
            "click_on_text is deprecated and will be removed in version 2. Use click_on_button instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.click_on_button(text, verify=verify)

    def click_on_button(self, button: str | int, verify: bool = True):
        """Simulate a click on an inline button by text or index.

        When `button` is a string, it searches for a button with matching text.
        When `button` is an integer, it treats it as the index of the button in
        the inline keyboard (flattened).
        """
        inline_keyboard = self.fake_bot_post.call_args[1]["payload"]["reply_markup"]["inline_keyboard"]
        if isinstance(button, str):
            callback_data = [item for row in inline_keyboard for item in row if item["text"] == button][0][
                "callback_data"
            ]
        elif isinstance(button, int):
            flat_buttons = [item for row in inline_keyboard for item in row]
            callback_data = flat_buttons[button]["callback_data"]
        else:
            raise ValueError("button must be a string or an integer index")

        data = self.construct_telegram_callback_query(callback_data)
        response = self.post_data(data, verify=verify)
        return response

    def send_text(self, text: str, verify: bool = True):
        """Simulate sending a text message."""
        payload = self.construct_telegram_update(text)
        return self.post_data(payload, verify=verify)

    def post_data(self, data: dict, verify: bool = True):
        """Post data to the webhook."""
        response = self.client.post(
            self.webhook_url,
            data=data,
            headers={"X-Telegram-Bot-Api-Secret-Token": settings.WEBHOOK_TOKEN},
            content_type="application/json",
        )
        if verify:
            self.assertEqual(response.json(), {"status": "ok", "message": "Message received."})
        return response

    @property
    def last_bot_message(self) -> str:
        """Return the last message sent by the bot.

        This is a convenience property for tests and can be used to verify the last message.

        Example:
            self.send_text("/start")
            self.assertEqual(self.last_bot_message, "Welcome to the bot!")
        """
        return self.fake_bot_post.call_args[1]["payload"]["text"]

    @staticmethod
    def construct_telegram_update(message_text: str):
        """Construct a minimal telegram update."""
        return {"message": {"chat": {"id": 123456789}, "text": message_text}}

    @staticmethod
    def construct_telegram_callback_query(callback_data: str):
        """Construct a minimal telegram callback query."""
        return {
            "callback_query": {
                "message": {"message_id": 123, "chat": {"id": 123456789, "first_name": "test", "type": "private"}},
                "data": callback_data,
            }
        }



"""Poll command for the sample bot."""

from django_telegram_app.bot import bot
from django_telegram_app.bot.base import BaseBotCommand, Step, TelegramUpdate


class Command(BaseBotCommand):
    """Poll command."""

    description = "Poll for a user's favourite sport."

    @property
    def steps(self):
        """Return the steps of the command."""
        return [AskFavouriteSport(self), Confirm(self), Respond(self)]


class AskFavouriteSport(Step):
    """Ask favourite sport step."""

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        data = self.get_callback_data(telegram_update)
        current_page: int = data.get("current_page", 1)
        start = (current_page - 1) * 3
        end = start + 3
        data.pop("favourite_sport", None)  # Remove any previous selection
        options = self.get_possible_answers()
        keyboard = [
            [{"text": text, "callback_data": self.next_step_callback(data, favourite_sport=value)}]
            for text, value in options[start:end]
        ]
        self._maybe_add_pagination_buttons(keyboard, options, data, current_page, end=end)
        bot.send_message(
            "What is your favourite sport?",
            self.command.settings.chat_id,
            reply_markup={"inline_keyboard": keyboard},
            message_id=telegram_update.message_id,
        )

    def get_possible_answers(self):
        """Get the possible options for the poll command."""
        return [
            ("🏓 Ping Pong", "Ping Pong"),
            ("🤺 Fencing", "Fencing"),
            ("🏸 Badminton", "Badminton"),
            ("🥊 Boxing", "Boxing"),
            ("🏹 Archery", "Archery"),
            ("🏒 Hockey", "Hockey"),
        ]

    def _maybe_add_pagination_buttons(self, keyboard: list, days: list, data: dict, current_page: int, end: int):
        if current_page > 1:
            back_callback = self.current_step_callback(data, current_page=current_page - 1)
            keyboard.append([{"text": "⬅️ Back", "callback_data": back_callback}])
        if len(days) > end:
            next_callback = self.current_step_callback(data, current_page=current_page + 1)
            keyboard.append([{"text": "➡️ Next", "callback_data": next_callback}])


class Confirm(Step):
    """Confirm step."""

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        data = self.get_callback_data(telegram_update)
        favourite_sport = data["favourite_sport"]
        callback_yes = self.next_step_callback(data, confirmed=True)
        callback_no = self.cancel_callback(
            data, confirmed=False, cancel_text="Poll cancelled. Your favourite sport was not recorded."
        )
        callback_previous = self.previous_step_callback(steps_back=1, original_data=data)
        keyboard = [
            [{"text": "✅ Yes", "callback_data": callback_yes}],
            [{"text": "❌ No", "callback_data": callback_no}],
            [{"text": "⬅️ Previous step", "callback_data": callback_previous}],
        ]
        bot.send_message(
            f"Would you like to submit {favourite_sport} as your favourite sport?",
            self.command.settings.chat_id,
            reply_markup={"inline_keyboard": keyboard},
            message_id=telegram_update.message_id,
        )


class Respond(Step):
    """Respond step."""

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        data = self.get_callback_data(telegram_update)
        favourite_sport = data["favourite_sport"]
        bot.send_message(
            f"Thank you! Your favourite sport {favourite_sport} has been recorded.",
            self.command.settings.chat_id,
            message_id=telegram_update.message_id,
        )
        self.command.next_step(self.name, telegram_update)



"""Tests to verify that django_telegram_app is configured correctly."""

import importlib

from django.test import SimpleTestCase


class AppsTests(SimpleTestCase):
    """Tests for django_telegram_app apps module."""

    def test_import_and_urls(self):
        """Test that the package can be imported and URLs are present."""
        import django_telegram_app  # noqa: F401

        mod = importlib.import_module("django_telegram_app.urls")
        assert hasattr(mod, "urlpatterns")



"""Telegram models."""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.translation import gettext_lazy as _

if TYPE_CHECKING:
    from django_telegram_app.bot.base import TelegramUpdate


class Message(models.Model):
    """Represent a Telegram Message.

    The raw message from the Telegram API is stored as json.

    Reference:
    https://core.telegram.org/bots/api#message
    """

    raw_message = models.JSONField(verbose_name=_("raw message"))
    error = models.TextField(verbose_name=_("error"), null=True, blank=True)

    @property
    def message_truncated(self):
        """Return the message truncated to 100 characters."""
        message_str = str(self.raw_message)
        if len(message_str) > 100:
            return message_str[:97] + "..."
        return message_str

    @property
    def update_id(self) -> int:
        """Return the chat id from the raw message."""
        return self.raw_message.get("update_id", "unknown")

    def __str__(self):
        """Return the string representation of the message."""
        if self.error:
            return f"{self.update_id} - {self.error}"
        return str(self.update_id)

    class Meta:
        """Set meta options."""

        verbose_name = _("message")
        verbose_name_plural = _("messages")


class AbstractTelegramSettings(models.Model):
    """Represent telegram settings."""

    if TYPE_CHECKING:
        data: models.JSONField[dict[str, str]]

    chat_id = models.IntegerField(verbose_name=_("chat id"), unique=True)
    data = models.JSONField(verbose_name=_("data"), default=dict, blank=True, encoder=DjangoJSONEncoder)
    updated_at = models.DateTimeField(verbose_name=_("updated at"), auto_now=True)

    class Meta:
        """Set meta options."""

        abstract = True

    @classmethod
    def create_from_telegram_update(cls, telegram_update: TelegramUpdate):
        """Create telegram settings from a telegram update.

        Subclasses can override this method to customize the creation process.
        """
        return cls.objects.create(chat_id=telegram_update.chat_id)

    def __str__(self):
        """Return a string representation of the telegram setting."""
        return f"Chat {self.chat_id}"


class TelegramSettings(AbstractTelegramSettings):
    """Concrete implementation of AbstractTelegramSettings."""

    class Meta:  # type: ignore[reportIncompatibleVariableOverride]
        """Set meta options."""

        verbose_name = _("telegram setting")
        verbose_name_plural = _("telegram settings")
        swappable = "TELEGRAM_SETTINGS_MODEL"


class CallbackData(models.Model):
    """Store callback data for Telegram inline keyboards.

    This is required because telegram limits the size of callback data to 64 bytes.
    """

    if TYPE_CHECKING:
        data: models.JSONField[dict[str, Any]]

    token = models.UUIDField(verbose_name=_("token"), default=uuid.uuid4, unique=True, db_index=True)
    command = models.CharField(verbose_name=_("command"), max_length=255)
    step = models.CharField(verbose_name=_("step"), max_length=255)
    action = models.CharField(
        verbose_name=_("action"),
        max_length=99,
        help_text=_("Name of a function on the command"),
    )
    data = models.JSONField(verbose_name=_("callback data"), default=dict, encoder=DjangoJSONEncoder)
    created_at = models.DateTimeField(verbose_name=_("created at"), auto_now_add=True)

    class Meta:
        """Set meta options."""

        verbose_name = _("callback data")
        verbose_name_plural = _("callback data")
        indexes = [models.Index(models.F("data__correlation_key"), name="callback_correlation_key_idx")]

    def __str__(self):
        """Return a string representation of the callback data."""
        return f"{self.token} - {self.data_truncated}"

    @property
    def data_truncated(self):
        """Return the data truncated to 100 characters."""
        data_str = str(self.data)
        if len(data_str) > 100:
            return data_str[:97] + "..."
        return data_str



"""Tests for the management package."""

from io import StringIO
from unittest.mock import patch

from django.core.management import call_command
from django.core.management.base import CommandError

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.bot.testing.testcases import TelegramBotTestCase

SETWEBHOOK_PATH = "django_telegram_app.management.commands.setwebhook"


class ManagementCommandTests(TelegramBotTestCase):
    """Tests for management commands."""

    @classmethod
    def setUpTestData(cls):
        """Set up test data."""
        cls.telegram_setting = get_telegram_settings_model().objects.create(chat_id=123456789)

    def test_base_management_command_runs(self):
        """Test that the BaseManagementCommand can be invoked without errors."""
        out = StringIO()
        call_command("poll", stdout=out)

    def test_base_management_command_checks_command_attribute(self):
        """Test that the command raises if the 'command' attribute is None."""
        # Patch samplebot's poll command to None to trigger ValueError
        out = StringIO()
        with self.assertRaises(ValueError):
            with patch("tests.testapps.samplebot.management.commands.poll.Command.command", None):
                call_command("poll", stdout=out)

    def test_base_management_command_respects_force_and_should_run(self):
        """Test that the --force option works as expected.

        force=False, should_run=False -> no output
        force=True, should_run=False -> output expected
        force=False, should_run=True -> output expected
        force=True, should_run=True -> output expected
        """
        out = StringIO()

        with patch("tests.testapps.samplebot.management.commands.poll.Command.should_run", return_value=False):
            call_command("poll", force=False, stdout=out)
            self.assertIn("Command '/poll' skipped as `should_run` returned False.", out.getvalue())
            call_command("poll", force=True, stdout=out)
            self.assertIn(f"Started the command for {self.telegram_setting}.", out.getvalue())

        with patch("tests.testapps.samplebot.management.commands.poll.Command.should_run", return_value=True):
            call_command("poll", force=False, stdout=out)
            self.assertIn(f"Started the command for {self.telegram_setting}.", out.getvalue())
            call_command("poll", force=True, stdout=out)
            self.assertIn(f"Started the command for {self.telegram_setting}.", out.getvalue())

        self.telegram_setting.delete()
        call_command("poll", force=True, stdout=out)
        self.assertIn("No Telegram-settings found for the given filter. Nothing to do.", out.getvalue())

    def test_set_webhook_command(self):
        """Test that the set_webhook command runs without errors."""
        out = StringIO()
        # Patch requests.post to avoid real HTTP calls
        with patch(f"{SETWEBHOOK_PATH}.requests.post") as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {"ok": True, "result": True}
            call_command("setwebhook", "https://example.com", stdout=out)
        self.assertIn("Successfully set webhook to", out.getvalue())

    def test_set_webhook_command_no_token(self):
        """Test that the set_webhook command runs without errors."""
        out = StringIO()
        # Patch requests.post to avoid real HTTP calls
        with patch(f"{SETWEBHOOK_PATH}.requests.post") as mock_post:
            with patch(f"{SETWEBHOOK_PATH}.app_settings.WEBHOOK_TOKEN", ""):
                mock_post.return_value.status_code = 200
                mock_post.return_value.json.return_value = {"ok": True, "result": True}
                call_command("setwebhook", "https://example.com", stdout=out)
        self.assertIn("Successfully set webhook to", out.getvalue())

    def test_set_webhook_command_failure(self):
        """Test that the set_webhook command handles failure correctly."""
        out = StringIO()
        err = StringIO()
        # Patch requests.post to simulate a failure response
        with patch(f"{SETWEBHOOK_PATH}.requests.post") as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {"ok": False, "description": "Invalid URL"}
            with self.assertRaises(CommandError, msg="Failed to set webhook to"):
                call_command("setwebhook", "https://example.com", stdout=out, stderr=err)
            self.assertEqual(out.getvalue(), "")
            self.assertIn("Something went wrong while setting the webhook", err.getvalue())



"""URLs for test project."""

from django.urls import include, path

from django_telegram_app.conf import settings as telegram_app_settings

urlpatterns = [
    path(telegram_app_settings.ROOT_URL, include("django_telegram_app.urls")),
]



"""Minimal Django settings for running tests."""

SECRET_KEY = "test-secret"
DEBUG = True
USE_TZ = True
ROOT_URLCONF = "tests.testapps.urls"

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_telegram_app",  # the app we are testing
    "tests.testapps.samplebot",  # tiny sample app
]

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# Config your package reads (from README Quickstart)
TELEGRAM = {
    "BOT_URL": "https://api.telegram.org/bot123:abc/",
    "ROOT_URL": "telegram/",
    "WEBHOOK_URL": "webhook",
    "WEBHOOK_TOKEN": "test-webhook-token",
}



"""Tests for django_telegram_app checks."""

from unittest.mock import patch

from django.apps import apps
from django.core import checks
from django.test import SimpleTestCase, override_settings


class ChecksTests(SimpleTestCase):
    """Tests for django_telegram_app checks."""

    @staticmethod
    def run_telegram_checks():
        """Run telegram system checks and return the results."""
        return checks.run_checks(app_configs=[apps.get_app_config("django_telegram_app")])

    def test_checks_run_without_errors(self):
        """Test that system checks run without errors."""
        errors = self.run_telegram_checks()
        self.assertEqual(errors, [])

    def test_check_telegram_required_settings_missing_settings(self):
        """Test that an error is returned when settings are missing."""
        from django_telegram_app.conf import settings

        with patch.object(settings, "missing_settings", return_value=["BOT_URL"]):
            errors = self.run_telegram_checks()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "telegram.E001")

    @override_settings(
        TELEGRAM_SETTINGS_MODEL="badformat",  # not app_label.ModelName
    )
    def test_check_swappable_telegram_settings_improperly_configured(self):
        """Test that an error is returned when TELEGRAM_SETTINGS_MODEL is improperly configured."""
        errors = self.run_telegram_checks()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "telegram.E002")

    def test_check_swappable_telegram_settings_unexpected_error(self):
        """Test that an error is returned when an unexpected error occurs during model resolution."""
        with patch("django_telegram_app.checks.get_telegram_settings_model") as mock_get_model:
            mock_get_model.side_effect = Exception("Simulated error")
            errors = self.run_telegram_checks()

        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "telegram.E003")

    @override_settings(
        TELEGRAM_SETTINGS_MODEL="auth.User",  # installed, but wrong base class
    )
    def test_check_swappable_telegram_settings_wrong_base_class(self):
        """Test that an error is returned when TELEGRAM_SETTINGS_MODEL does not subclass AbstractTelegramSettings."""
        errors = self.run_telegram_checks()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "telegram.E004")

    def test_check_get_commands_unexpected_error(self):
        """Test that an error is returned when an unexpected error occurs during command discovery."""
        with patch("django_telegram_app.checks.get_commands") as mock_get_commands:
            mock_get_commands.side_effect = Exception("Simulated error")
            errors = self.run_telegram_checks()

        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "telegram.E005")



"""Sample app configuration."""

from django.apps import AppConfig


class SampleBotConfig(AppConfig):
    """Sample bot app config."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "tests.testapps.samplebot"
    label = "tests_samplebot"



"""URL configuration for telegram app."""

from django.urls import path

from django_telegram_app import views
from django_telegram_app.conf import settings

urlpatterns = [
    path(settings.WEBHOOK_URL, views.webhook, name="webhook"),
]



"""Telegram app configuration."""

from django.apps import AppConfig


class TelegramConfig(AppConfig):
    """Represent the telegram AppConfig."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_telegram_app"
    label = "django_telegram_app"

    def ready(self):
        """Import checks."""
        import django_telegram_app.checks  # noqa: F401



"""Base classes for writing telegrambot commands.

telegrambot commands are named commands which can execute operations in response to user messages.
"""

from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from django_telegram_app.models import CallbackData

if TYPE_CHECKING:
    from django_telegram_app.models import AbstractTelegramSettings


class BaseBotCommand:
    """Represent a base Telegram bot command.

    This is the base class for all user-defined Telegram bot commands.
    """

    description: str = ""

    def __init__(self, settings: AbstractTelegramSettings):
        """Initialize the command."""
        self.settings = settings

    def start(self, telegram_update: TelegramUpdate):
        """Start the command."""
        logging.info(f"Starting {self.get_name()} for {self.settings}")
        self._clear_state()
        return self.steps[0].handle(telegram_update)

    def finish(self, current_step_name: str, telegram_update: TelegramUpdate):
        """Finish the command and clear all data."""
        logging.info(f"Finishing the command at step {current_step_name}")
        self._clear_state()
        self._clear_callback_data(telegram_update)

    def cancel(self, current_step_name: str, telegram_update: TelegramUpdate):
        """Cancel the command and clear all data."""
        from django_telegram_app.bot.bot import send_message

        logging.info(f"Canceled the command at step {current_step_name}")
        data = self.get_callback_data(telegram_update.callback_data)
        cancel_text = data.get("cancel_text", "Command canceled.")
        send_message(cancel_text, self.settings.chat_id)
        return self.finish(current_step_name, telegram_update)

    def next_step(self, current_step_name: str, telegram_update: TelegramUpdate):
        """Proceed to the next step in the command."""
        next_index = self._steps_to_str().index(current_step_name) + 1
        if next_index < len(self.steps):
            next_step = self.steps[next_index]
            return next_step.handle(telegram_update)
        self.finish(current_step_name, telegram_update)

    def previous_step(self, current_step_name: str, telegram_update: TelegramUpdate):
        """Return to the previous step in the command."""
        data = self.get_callback_data(telegram_update.callback_data)
        steps_back = int(data.get("_steps_back", 1))
        previous_index = self._steps_to_str().index(current_step_name) - steps_back
        if previous_index >= 0:
            previous_step = self.steps[previous_index]
            return previous_step.handle(telegram_update)

    def current_step(self, current_step_name: str, telegram_update: TelegramUpdate):
        """Reload the current step."""
        current_index = self._steps_to_str().index(current_step_name)
        current_step = self.steps[current_index]
        return current_step.handle(telegram_update)

    def create_callback(self, step_name: str, action: str, **kwargs):
        """Create callback data for the current command and return the token."""
        if not kwargs:
            kwargs = self._get_default_callback_data()
        if "correlation_key" not in kwargs:
            kwargs.update(self._get_default_callback_data())
        callback_data = CallbackData(command=self.get_command_string(), step=step_name, action=action, data=kwargs)
        callback_data.save()
        return str(callback_data.token)

    def get_callback(self, token: str):
        """Return the callback for the given token."""
        return CallbackData.objects.get(token=token)

    def get_callback_data(self, callback_token: str) -> dict[str, Any]:
        """Get callback data from the callback token.

        If the callback token is not provided, return default callback data.
        """
        if not callback_token:
            return self._get_default_callback_data()
        callback_data = self.get_callback(callback_token)
        return callback_data.data

    @property
    def steps(self) -> Sequence[Step]:
        """Return the steps of the command."""
        raise NotImplementedError("Subclasses must implement this method")

    @classmethod
    def get_name(cls):
        """Return the name of the command.

        By default this is the lowercased last part of the module name.
        """
        return cls.__module__.split(".")[-1].lower()

    @classmethod
    def get_command_string(cls):
        """Return the command string."""
        return f"/{cls.get_name()}"

    def _get_default_callback_data(self):
        """Return a dictionary with correlation key as default callback data."""
        return {"correlation_key": str(uuid.uuid4())}

    def _clear_state(self):
        """Clear the command state."""
        self.settings.data = {}
        self.settings.save()

    def _clear_callback_data(self, telegram_update: TelegramUpdate):
        """Clear callback data for the current command."""
        step_data = self.get_callback_data(telegram_update.callback_data)
        correlation_key = step_data.get("correlation_key", "non_existent_key")
        CallbackData.objects.filter(data__correlation_key=correlation_key).delete()

    def _steps_to_str(self):
        return [step.name for step in self.steps]


class Step:
    """Represent a step in a Telegram bot command.

    This is the base class for all user-defined steps.
    """

    def __init__(self, command: BaseBotCommand, unique_id: str | None = None):
        """Initialize the step."""
        self.command = command
        self.unique_id = unique_id

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        raise NotImplementedError("This method should be overridden by subclasses.")

    def next_step_callback(self, original_data: dict | None = None, **kwargs):
        """Create a callback to advance to the next step."""
        return self._create_callback("next_step", original_data, **kwargs)

    def previous_step_callback(self, steps_back: int, original_data: dict | None = None, **kwargs):
        """Create a callback to return to the previous step."""
        kwargs["_steps_back"] = steps_back
        return self._create_callback("previous_step", original_data, **kwargs)

    def current_step_callback(self, original_data: dict | None = None, **kwargs):
        """Create a callback to reload the current step with the provided data."""
        return self._create_callback("current_step", original_data, **kwargs)

    def cancel_callback(self, original_data: dict | None = None, **kwargs):
        """Create a callback to cancel the command."""
        return self._create_callback("cancel", original_data, **kwargs)

    def get_callback_data(self, telegram_update: TelegramUpdate):
        """Get callback data from the telegram_update.

        If the update is a message and not a command, check if we are waiting for user input.
        If so, retrieve the callback data using the waiting_for token and store the message text
        in the appropriate key in the callback data.

        Otherwise, retrieve the callback data using the callback token from the update.
        If no callback token is provided, return default callback data.
        """
        if not telegram_update.callback_data and telegram_update.is_message() and not telegram_update.is_command():
            waiting_for = self.command.settings.data.get("_waiting_for", None)
            if waiting_for:
                callback_token = waiting_for
                callback_data = self.command.get_callback_data(callback_token)
                key = callback_data["_message_key"]  # Move the message_text to this key
                callback_data[key] = telegram_update.message_text.strip()
                return callback_data

        callback_token = telegram_update.callback_data
        return self.command.get_callback_data(callback_token)

    def add_waiting_for(self, message_key: str, data: dict[str, Any] | None = None):
        """Add waiting_for to the command settings.

        The message_key will be used to store the user input in the callback data of the next step.
        """
        data = data or {}
        self.command.settings.data["_waiting_for"] = self.next_step_callback(data, _message_key=message_key)
        self.command.settings.save()

    @property
    def name(self):
        """Return the name of the step."""
        return self.unique_id or type(self).__name__

    def _create_callback(self, action: str, original_data: dict | None = None, **kwargs):
        """Create callback data for the current step and return the token."""
        original_data = original_data or {}
        data = {**original_data, **kwargs}
        return self.command.create_callback(self.name, action, **data)


class TelegramUpdate:
    """Represent a normalized Telegram update."""

    def __init__(self, update: dict):
        """Initialize the normalized Telegram update."""
        self.message = cast(dict | None, update.get("message"))
        self.callback_query = cast(dict | None, update.get("callback_query"))

        if self.message and "text" in self.message:
            self.chat_id = int(self.message["chat"]["id"])
            self.message_id = 0
            self.message_text = str(self.message["text"])
            self.callback_data = ""
        elif self.callback_query:
            self.chat_id = int(self.callback_query["message"]["chat"]["id"])
            self.message_id = int(self.callback_query["message"]["message_id"])
            self.message_text = ""
            self.callback_data = str(self.callback_query.get("data"))
        else:
            raise ValueError("Unsupported Telegram update format")

    def is_message(self):
        """Return the message part of the update."""
        return self.message is not None

    def is_callback_query(self):
        """Return the callback query part of the update."""
        return self.callback_query is not None

    def is_command(self):
        """Check if the update is a command."""
        return self.is_message() and self.message_text.startswith("/")



"""Tests for the bot package."""

import uuid
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

from django.test import SimpleTestCase

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.bot import get_commands, load_command_class
from django_telegram_app.bot.testing.testcases import TelegramBotTestCase
from django_telegram_app.models import CallbackData, Message
from tests.testapps.samplebot.telegrambot.commands.echo import Command as EchoCommand
from tests.testapps.samplebot.telegrambot.commands.poll import Command as PollCommand


class BotTests(TelegramBotTestCase):
    """Tests for the bot package."""

    @classmethod
    def setUpTestData(cls):
        """Set up test data."""
        cls.telegram_setting = get_telegram_settings_model().objects.create(chat_id=123456789)

    def test_get_commands_returns_empty_dict_when_not_configured(self):
        """Test that get_commands returns an empty dict when settings are not configured."""
        get_commands.cache_clear()
        fake_settings = SimpleNamespace(configured=False)
        with patch("django_telegram_app.bot.settings", fake_settings):
            commands = get_commands()
        self.assertEqual(commands, {})

        # Ensure cache is cleared for other tests, otherwise none would find commands
        get_commands.cache_clear()

    def test_discovery_finds_poll_and_echo(self):
        """Test that the poll and echo commands are discovered and can be loaded."""
        expected_commands = {"poll": PollCommand, "echo": EchoCommand}
        for cmd, expected_class in expected_commands.items():
            assert cmd in get_commands().keys()

            appname = get_commands()[cmd]
            cmd_instance = load_command_class(appname, cmd, self.telegram_setting)
            assert isinstance(cmd_instance, expected_class)
            assert cmd_instance.get_command_string() == f"/{cmd}"

    def test_telegram_invalid_token(self):
        """Test the telegram app with an invalid token."""
        response = self.client.post(
            self.webhook_url,
            data={},
            headers={"X-Telegram-Bot-Api-Secret-Token": "invalid_token"},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"status": "error", "message": "Invalid token."})

    def test_telegram_send_help(self):
        """Test the telegram send help command.

        The help text should always be displayed when an unknown command/message/option is sent.
        """
        self.send_text("dummy text")
        self.assertEqual(self.fake_bot_post.call_count, 1)
        self.assertEqual(self.fake_bot_post.call_args.args[0], "sendMessage")
        self.assertIn("Currently available commands", self.fake_bot_post.call_args[1]["payload"]["text"])

    def test_poll_command(self):
        """Test the poll command."""
        self.send_text("/poll")
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button("🏓 Ping Pong")
        self.assertEqual(self.last_bot_message, "Would you like to submit Ping Pong as your favourite sport?")
        self.click_on_button(0)  # First button is "✅ Yes"
        self.assertEqual(self.last_bot_message, "Thank you! Your favourite sport Ping Pong has been recorded.")

    def test_poll_command_next_page(self):
        """Test the next page functionality in the poll command."""
        self.send_text("/poll")
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button(-1)  # Last button is "➡️ Next page"
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button("🥊 Boxing")
        self.assertRegex(self.last_bot_message, "Would you like to submit Boxing as your favourite sport?")
        self.click_on_button("✅ Yes")
        self.assertEqual(self.last_bot_message, "Thank you! Your favourite sport Boxing has been recorded.")

    def test_poll_command_cancel(self):
        """Test cancelling the poll command."""
        self.send_text("/poll")
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button("🤺 Fencing")
        self.assertEqual(self.last_bot_message, "Would you like to submit Fencing as your favourite sport?")
        self.click_on_button("❌ No")
        self.assertEqual(self.last_bot_message, "Poll cancelled. Your favourite sport was not recorded.")

    def test_poll_command_previous(self):
        """Test using the previous button in the poll command."""
        self.send_text("/poll")
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button("🏸 Badminton")
        self.assertEqual(self.last_bot_message, "Would you like to submit Badminton as your favourite sport?")
        self.click_on_button("⬅️ Previous step")
        self.assertEqual(self.last_bot_message, "What is your favourite sport?")
        self.click_on_button("🤺 Fencing")
        self.assertEqual(self.last_bot_message, "Would you like to submit Fencing as your favourite sport?")
        self.click_on_button("✅ Yes")
        self.assertEqual(self.last_bot_message, "Thank you! Your favourite sport Fencing has been recorded.")

    def test_echo_command(self):
        """Test the echo command."""
        self.send_text("/echo")
        self.assertEqual(self.last_bot_message, "Send the message you want to echo:")
        self.send_text("Hello, World!")
        self.assertEqual(self.last_bot_message, "You said: Hello, World!")

    def test_unknown_command(self):
        """Test sending an unknown command."""
        self.send_text("/unknowncommand")
        self.assertIn("Currently available commands", self.last_bot_message)

    def test_unexpected_error(self):
        """Test handling an unexpected error in a command."""
        with patch("django_telegram_app.bot.bot.handle_update", MagicMock()) as fake_bot_handle_update:
            fake_bot_handle_update.side_effect = Exception("Simulated error")
            response = self.send_text("/whatever", verify=False)
        self.assertEqual(response.json(), {"status": "error", "message": "Message received."})
        last_message = Message.objects.last()
        assert last_message is not None  # Use assertion to satisfy type checker
        assert last_message.error is not None  # Use assertion to satisfy type checker
        self.assertIn("Simulated error", last_message.error)

    def test_token_is_valid_if_not_configured(self):
        """Test that any token is valid if BOT_API_SECRET_TOKEN is not configured."""
        with patch("django_telegram_app.bot.bot.settings.WEBHOOK_TOKEN", ""):
            response = self.client.post(
                self.webhook_url,
                data={},
                headers={"X-Telegram-Bot-Api-Secret-Token": "any_token"},
                content_type="application/json",
            )
        self.assertEqual(response.status_code, 200)

    def test_call_command_step_do_nothing(self):
        """Test that calling a command step with token DO_NOTHING it does nothing."""
        from django_telegram_app.bot.bot import DO_NOTHING, _call_command_step

        called = _call_command_step(DO_NOTHING, MagicMock(), MagicMock())
        self.assertFalse(called)

    def test_call_command_step_callback_not_found(self):
        """Test that calling a command step with an invalid token raises."""
        from django_telegram_app.bot.bot import _call_command_step

        uuid_token = str(uuid.uuid4())
        CallbackData.objects.filter(token=uuid_token).delete()  # Ensure it does not exist
        called = _call_command_step(uuid_token, MagicMock(), MagicMock())
        self.assertFalse(called)

    def test_command_has_steps_abstract(self):
        """Test that accessing steps property on BaseBotCommand raises NotImplementedError."""
        from django_telegram_app.bot.base import BaseBotCommand

        command = BaseBotCommand(MagicMock())
        with self.assertRaises(NotImplementedError):
            _ = command.steps

    def test_step_handle_abstract(self):
        """Test that calling handle method on Step raises NotImplementedError."""
        from django_telegram_app.bot.base import Step

        step = Step(MagicMock())
        with self.assertRaises(NotImplementedError):
            step.handle(MagicMock())

    def test_steps_back_does_nothing_when_idx_out_of_bounds(self):
        """Test that steps_back does nothing when index would go out of bounds."""
        from django_telegram_app.bot.base import BaseBotCommand, Step

        class DummyCommand(BaseBotCommand):
            @property
            def steps(self):
                return [Step(self, unique_id="step_1"), Step(self, unique_id="step_2")]

        telegram_settings = MagicMock(name="telegram_settings")
        telegram_update = MagicMock(name="telegram_update")

        command = DummyCommand(telegram_settings)
        with patch.object(command, "get_callback_data", return_value={"_steps_back": 1}):
            command.previous_step("step_1", telegram_update)  # Go back from first step, this should just do nothing

    def test_step_get_callback_data_calls_command_get_callback_data_if_not_waiting_for_input(self):
        """Test that Step.get_callback_data calls Command.get_callback_data if not waiting for input."""
        from django_telegram_app.bot.base import BaseBotCommand, Step

        class DummyCommand(BaseBotCommand):
            @property
            def steps(self):
                return [Step(self, unique_id="step_1"), Step(self, unique_id="step_2")]

        telegram_settings = MagicMock(name="telegram_settings")
        telegram_update = MagicMock(name="telegram_update")
        telegram_update.callback_data = ""
        telegram_update.is_message.return_value = True
        telegram_update.is_command.return_value = False
        telegram_settings.data = {}
        command = DummyCommand(telegram_settings)
        with patch.object(command, "get_callback_data") as fake_get_callback_data:
            command.steps[0].get_callback_data(telegram_update)
            fake_get_callback_data.assert_called_once_with(telegram_update.callback_data)

    def test_get_or_create_telegram_settings_creates_if_allowed(self):
        """Test that get_or_create_telegram_settings creates settings if allowed."""
        from django_telegram_app.bot.bot import _get_or_create_telegram_settings
        from django_telegram_app.conf import settings

        telegram_update = SimpleNamespace(chat_id=987654321)
        with patch.object(settings, "ALLOW_SETTINGS_CREATION_FROM_UPDATES", False):
            with self.assertRaises(get_telegram_settings_model().DoesNotExist):
                _get_or_create_telegram_settings(telegram_update)  # type: ignore[reportArgumentType]

        with patch.object(settings, "ALLOW_SETTINGS_CREATION_FROM_UPDATES", True):
            telegram_settings = _get_or_create_telegram_settings(telegram_update)  # type: ignore[reportArgumentType]
            self.assertIsNotNone(telegram_settings)
            self.assertEqual(telegram_settings.chat_id, 987654321)

    def test_create_callback_provides_default(self):
        """Test that create_callback provides a default value if no kwargs are provided."""
        from django_telegram_app.bot.base import BaseBotCommand

        telegram_settings = MagicMock(name="telegram_settings")
        command = BaseBotCommand(telegram_settings)
        callback_token = command.create_callback("dummy_step", "next_step")
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

    def test_step_create_callback_always_includes_correlation_key(self):
        """Test that Step.*_step_callback provides includes a correlation key if no or bad original data is provided."""
        from django_telegram_app.bot.base import BaseBotCommand, Step

        telegram_settings = MagicMock(name="telegram_settings")
        command = BaseBotCommand(telegram_settings)
        step = Step(command)
        callback_token = step.next_step_callback(some_value=123)
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

        callback_token = step.previous_step_callback(1, some_value=123)
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

        callback_token = step.current_step_callback(some_value=123)
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

        callback_token = step.cancel_callback(some_value=123)
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

        callback_token = step.next_step_callback(original_data={"faulty": "original"}, some_value=123)
        callback_data = command.get_callback(callback_token)
        self.assertIn("correlation_key", callback_data.data)

    def test_click_on_text_deprecation(self):
        """Test that click_on_text raises a deprecation warning."""
        with self.assertWarns(DeprecationWarning) as cm:
            with patch.object(self, "click_on_button", return_value=MagicMock()):
                self.click_on_text("Some Text")
        self.assertIn("click_on_text is deprecated", str(cm.warning))

    def test_click_on_button_by_index_invalid_type(self):
        """Test that click_on_button raises ValueError when given an invalid type."""
        with self.assertRaises(ValueError) as cm:
            with patch.object(self.fake_bot_post, "call_args", new=MagicMock()):
                self.click_on_button(3.14)  # Invalid type: float  # type: ignore[reportArgumentType]
        self.assertIn("button must be a string or an integer index", str(cm.exception))


class ExtraBotTests(SimpleTestCase):
    """Extra tests for bot functions which are mocked in BotTests."""

    def test_post(self):
        """Test the post function sends a request to the correct endpoint."""
        from django_telegram_app.bot.bot import post

        with patch("django_telegram_app.bot.bot.requests.post") as fake_requests_post:
            endpoint = "sendMessage"
            payload = {"chat_id": 123456789, "text": "Hello"}
            post(endpoint, payload)

            fake_requests_post.assert_called_once_with(
                "https://api.telegram.org/bot123:abc/sendMessage", json=payload, timeout=5
            )



"""Tests for the models package."""

from django.test import TestCase

from django_telegram_app.models import CallbackData, Message, TelegramSettings


class ModelsTests(TestCase):
    """Tests for the models package."""

    def test_message_message_truncated(self):
        """Test that the message_truncated property works as expected."""
        msg = Message(raw_message="a" * 500)
        self.assertEqual(len(msg.message_truncated), 100)
        self.assertEqual(msg.message_truncated[-3:], "...")

        msg_short = Message(raw_message="Short message")
        self.assertEqual(msg_short.message_truncated, "Short message")

    def test_message_update_id(self):
        """Test that the update_id property works as expected."""
        msg = Message(raw_message={"update_id": 12345})
        self.assertEqual(msg.update_id, 12345)

        msg_no_id = Message(raw_message={})
        self.assertEqual(msg_no_id.update_id, "unknown")

    def test_message_str(self):
        """Test that the __str__ method of Message works as expected."""
        msg = Message(raw_message={"update_id": 12345})
        self.assertEqual(str(msg), "12345")

        msg_with_error = Message(raw_message={"update_id": 67890}, error="Some error")
        self.assertEqual(str(msg_with_error), "67890 - Some error")

    def test_telegram_settings_str(self):
        """Test that the __str__ method of TelegramSettings works as expected."""
        settings = TelegramSettings(chat_id="67890")
        self.assertEqual(str(settings), "Chat 67890")

    def test_callback_data_str(self):
        """Test that the __str__ method of CallbackData works as expected."""
        uuid_token = "9e265e02-6b4c-41f8-8edd-c12e8e601469"
        data = {"key": "value"}
        callback = CallbackData(token=uuid_token, command="/test", step="step1", action="cancel", data=data)
        self.assertEqual(str(callback), f"{uuid_token} - {data}")

    def test_callback_data_data_truncated(self):
        """Test that the data_truncated property of CallbackData works as expected."""
        data = {"a" * 111: "value"}
        callback = CallbackData(command="/test", step="step1", action="cancel", data=data)
        self.assertEqual(len(callback.data_truncated), 100)
        self.assertEqual(callback.data_truncated[-3:], "...")
